# -------------------------------------------------------------------------------------------------
# test_ctypes_mp_xi2.py: test of ctypes in multiprocessing                              
#
# Theory:      Multiprocessing
#
# Description: creating shared ctypes objects from shared memory with the normal ctypes syntax 
#
# class multiprocessing.Process([group[, target[, name[, args[, kwargs]]]]]) 
# 
#    Process objects represent activity that is run in a separate process. 
#    The Process class has equivalents of all the methods of threading.Thread.
#    
#constructor
#    should always be called with keyword arguments. 
#    
#group
#    should always be None; it exists solely for compatibility with Thread. 
#    
#target
#    is the callable object to be invoked by the run() method. 
#    It defaults to None, meaning nothing is called. 
#    
#name
#    is the process name
#    by default, a unique name is constructed of the form Process-N1:N2:...:Nk
#    where N1,N2,...,Nk is a sequence of integers whose length is determined by the generation of the process. 
#    
#args
#    is the argument tuple for the target invocation
#    
#kwargs
#    is a dictionary of keyword arguments for the target invocation. By default, no arguments are passed to target.
#    
#If a subclass overrides the constructor, it must make sure it invokes the base class constructor (Process.__init__()) before doing anything else to the process.
#    
#run() 
#    Method representing the process activity.
#    You may override this method in a subclass. 
#    The standard run() method invokes the callable object passed to the object constructor as the target argument, if any, 
#    with sequential and keyword arguments taken from the args and kwargs arguments, respectively.
#    
#start() 
#    Start the process activity.
#    This must be called at most once per process object. 
#    It arranges for the object run() method to be invoked in a separate process.
#    
#join([timeout]) 
#    Block the calling thread until the process whose join() method is called terminates or until the optional timeout occurs.
#    If timeout is None then there is no timeout.
#    A process can be joined many times.
#    A process cannot join itself because this would cause a deadlock. It is an error to attempt to join a process before it has been started.
#    
#name 
#    Return the process name.
#    
#Process.name = name 
#    Set the process name.
#    The name is a string used for identification purposes only. 
#    It has no semantics. 
#    Multiple processes may be given the same name. 
#    The initial name is set by the constructor.
#    
#is_alive() <--- OK
#    Return whether the process is alive.
#    Roughly, a process object is alive from the moment the start() method returns until the child process terminates.
#    
#daemon 
#    Return the process daemon flag., this is a boolean.
#    
#Process.daemon = daemonic 
#    Set the process daemon flag to the Boolean value daemonic. 
#    This must be called before start() is called.
#    The initial value is inherited from the creating process.
#    When a process exits, it attempts to terminate all of its daemonic child processes.
#    Note that a daemonic process is not allowed to create child processes. 
#    Otherwise a daemonic process would leave its children orphaned if it gets terminated when its parent process exits.
#        
#In addition process objects also support the following methods:
#    
#pid 
#    Return the process ID. Before the process is spawned, this will be None.
#    
#exitcode 
#    Return the child exit code. This will be None if the process has not yet terminated. A negative value -N indicates that the child was terminated by signal N.
#
#authkey 
#    Return the process authentication key (a byte string).
#    When multiprocessing is initialized the main process is assigned a random string using os.random().
#    When a Process object is created, it will inherit the authentication key of its parent process, although this may be changed using Process.authkey below.
#    See Authentication keys.
# 
#Process.authkey = authkey 
# Set the process authentication key which must be a byte string.
# 
#terminate() 
#    Terminate the process. 
#    On Unix this is done using the SIGTERM signal; 
#    on Windows TerminateProcess is used. Note that exit handlers and finally clauses, etc., will not be executed.
#    Note that descendant processes of the process will not be terminated, they will simply become orphaned
#    Warning
#    If this method is used when the associated process is using a pipe or queue 
#    then the pipe or queue is liable to become corrupted and may become unusable by other process. 
#    Similarly, if the process has acquired a lock or semaphore etc. then terminating it is liable to cause other processes to deadlock.
#    Note that the start(), join(), is_alive() and get_exit_code() methods should only be called by the process that created the process object.
# -------------------------------------------------------------------------------------------------

# -------------------------------------------------------------------------------------------------
# Imports from standard modules                              
# -------------------------------------------------------------------------------------------------
import os
import sys
from ctypes import Structure, c_double

# -------------------------------------------------------------------------------------------------
# Imports from multiprocessing package                              
# -------------------------------------------------------------------------------------------------
from multiprocessing import Process, Queue, Pipe, Lock, Manager, Pool
from multiprocessing.sharedctypes import Value, Array


# -------------------------------------------------------------------------------------------------
# yourDragonXi imports                               
# -------------------------------------------------------------------------------------------------
from process2xi import Process2Xi
from queue2xi  import Queue2Xi
from pipe2xi import Pipe2Xi
from lock2xi import Lock2Xi
from pool2xi import Pool2Xi
from value2xi import Value2Xi
from arrayfunction2xi import ArrayFunction2Xi #<--- note: this must be called (Array2Xi uses the class from managers.py)
from manager2xi import Manager2Xi 


# -------------------------------------------------------------------------------------------------
# Point2Xi
# -------------------------------------------------------------------------------------------------
class Point2Xi(Structure):
    _fields_ = [('x', c_double), ('y', c_double)]

 
# -------------------------------------------------------------------------------------------------
# modify2xi function
#    
def modify2xi(n_param, x_param, s_param, A_param):
    n_xi2 = n_param
    x_xi2 = x_param
    s_xi2 = s_param
    A_xi2 = A_param

    n_xi2.value **= 2
    x_xi2.value **= 2
    s_xi2.value = s_xi2.value.upper()
    
    for a_xi2 in A_xi2:
        a_xi2.x **= 2
        a_xi2.y **= 2

# -------------------------------------------------------------------------------------------------
# ThreeDimension2Xi
# -------------------------------------------------------------------------------------------------
class ThreeDimension3Xi(Structure):
    _fields_ = [('x', c_double), ('y', c_double), ('z', c_double)]  
        
# -------------------------------------------------------------------------------------------------
# modify3xyz function
#    
def modify3xi(n_param, x_param, s_param, A_param):
    n_xi3 = n_param
    x_xi3 = x_param
    s_xi3 = s_param
    A_xi3 = A_param

    n_xi3.value **= 3
    x_xi3.value **= 3
    s_xi3.value = s_xi3.value.upper()
    
    for a in A_xi3:
        a.x **= 3
        a.y **= 3
        a.z **= 3


# -------------------------------------------------------------------------------------------------
# sense function for multiprocessing
#
def sense(target_param, queue_param, pipe_param,lock_param,dict_param,list_param):
    target_xi2 = target_param
    queue_xi2 = queue_param
    pipe_publisher_channel_xi2 = pipe_param
    lock_xi2 = lock_param
    dict_xi2 = dict_param
    list_xi2 = list_param
    
    
    print("SENSE: ", target_xi2)
    # ---------------------------------------------------------------------------------------------
    # put a message to the queue
    # 
    queue_xi2.put(['DragonXi2_00', target_xi2, 'sensing!']) # puts to its queue
    # ---------------------------------------------------------------------------------------------
    # reserve the pipe using the lock, send messages and release the pipe
    # 
    lock_xi2.acquire()

    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_01', 'BIO ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_02', 'CO2 ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_03', 'Coal ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_04', 'Diesel ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_05', 'Electricity ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_06', 'Fusion ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_07', 'Gas ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_08', 'Geo ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_09', 'Hydrogen ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_10', 'Lithium ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_11', 'Nuclear Power ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_12', 'Ocean ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_13', 'Oil ',target_xi2]) 
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_14', 'Solar ',target_xi2])
    pipe_publisher_channel_xi2.send(['DragonXi2_Sensing_15', 'Wind ',target_xi2])
    
    dict_xi2[1] = "BIO"
    dict_xi2[2] = "CO2"
    dict_xi2[3] = "Coal"
    dict_xi2[4] = "Diesel"
    dict_xi2[5] = "Electricity"
    dict_xi2[6] = "Fusion"
    dict_xi2[7] = "Gas"
    dict_xi2[8] = "Geo"
    dict_xi2[9] = "Hydrogen"
    dict_xi2[10] = "Lithium"
    dict_xi2[11] = "Nuclear Power"
    dict_xi2[12] = "Ocean"
    dict_xi2[13] = "Oil"
    dict_xi2[14] = "Solar"
    dict_xi2[15] = "Wind"
    

    lock_xi2.release()       
# -------------------------------------------------------------------------------------------------
# reason funcion for multiprocessing
# 
def reason(target_param, queue_param, pipe_param,lock_param,dict_param,list_param):
    target_xi4 = target_param
    queue_xi4 = queue_param
    pipe_publisher_channel_xi4 = pipe_param
    lock_xi4 = lock_param
    dict_xi4 = dict_param
    list_xi8 = list_param
    
    print("REASON: ", target_xi4)
    # ---------------------------------------------------------------------------------------------
    # put a message to the queue
    # 
    queue_xi4.put(['DragonXi4_00', target_xi4, 'portfolio must be sustainable!']) # puts to its queue 
    # ---------------------------------------------------------------------------------------------
    # reserve the pipe using the lock, send messages and release the pipe
    # 
    lock_xi4.acquire()
    pipe_publisher_channel_xi4.send(['DragonXi4_Reasoning_01', 'BIO ',target_xi4])
    pipe_publisher_channel_xi4.send(['DragonXi4_Reasoning_02', 'Geo ',target_xi4])
    pipe_publisher_channel_xi4.send(['DragonXi4_Reasoning_03', 'Hydrogen ',target_xi4])
    pipe_publisher_channel_xi4.send(['DragonXi4_Reasoning_04', 'Nuclear Power ',target_xi4])
    pipe_publisher_channel_xi4.send(['DragonXi4_Reasoning_05', 'Ocean ',target_xi4])
    pipe_publisher_channel_xi4.send(['DragonXi4_Reasoning_06', 'Solar ',target_xi4])
    pipe_publisher_channel_xi4.send(['DragonXi4_Reasoning_07', 'Wind ',target_xi4])    

    dict_xi4[1] = "BIO"
    dict_xi4[2] = "Geo"
    dict_xi4[3] = "Hydrogen"
    dict_xi4[4] = "Nuclear Power"
    dict_xi4[5] = "Ocean"
    dict_xi4[6] = "Solar"
    dict_xi4[7] = "Wind"
    
    lock_xi4.release()
        
# -------------------------------------------------------------------------------------------------
# ace function for multiprocessing
#
def ace(target_param, queue_param, pipe_param,lock_param,dict_param,list_param):
    # ---------------------------------------------------------------------------------------------
    # 
    # 
    target_xi8 = target_param
    queue_xi8 = queue_param
    pipe_publisher_channel_xi8 = pipe_param
    lock_xi8 = lock_param
    dict_xi8 = dict_param
    list_xi8 = list_param
    
    print("ACE: ", target_xi8)
    # ---------------------------------------------------------------------------------------------
    # put a message to the queue
    # 
    queue_xi8.put(['DragonXi8_00', target_xi8, 'to be aced!']) # puts to its queue       
    
    # ---------------------------------------------------------------------------------------------
    # reserve the pipe using the lock, send messages and release the pipe
    # 
    lock_xi8.acquire()
    pipe_publisher_channel_xi8.send(['DragonXi8_Ace_01', 'BIO ',target_xi8])
    pipe_publisher_channel_xi8.send(['DragonXi8_Ace_02', 'Hydrogen ',target_xi8])
    pipe_publisher_channel_xi8.send(['DragonXi8_Ace_03', 'Nuclear Power ',target_xi8])
    pipe_publisher_channel_xi8.send(['DragonXi8_Ace_04', 'Solar ',target_xi8])
    pipe_publisher_channel_xi8.send(['DragonXi8_Ace_05', 'Wind ',target_xi8])    

    dict_xi8[1] = "BIO"
    dict_xi8[2] = "Hydrogen"
    dict_xi8[3] = "Nuclear Power"
    dict_xi8[4] = "Solar"
    dict_xi8[5] = "Wind"
    
    lock_xi8.release()
# -------------------------------------------------------------------------------------------------
# expxi2 function for pool workers
#
def expxi2(x_param):
    xi2 = x_param
    return xi2*xi2

# -------------------------------------------------------------------------------------------------
# expxi4 function for pool workers
#
def expxi4(x_param):
    xi4 = x_param
    return xi4*xi4*xi4*xi4
    
# -------------------------------------------------------------------------------------------------
# expxi8 function for pool workers
#
def expxi8(x_param):
    xi8 = x_param
    return xi8*xi8*xi8*xi8*xi8*xi8*xi8*xi8         
    
# -------------------------------------------------------------------------------------------------
# phase1: original multiprocessing package used                              
# -------------------------------------------------------------------------------------------------
def phase1():
    # ---------------------------------------------------------------------------------------------
    # create multiprocessing queues for sensing, reasoning and acing
    # 
    queue2xi = Queue()
    queue4xi = Queue()
    queue8xi = Queue()
    # ---------------------------------------------------------------------------------------------
    # create multiprocessing pipes with consumer and publisher ends
    #     
    pipe2xi_consumer_channel, pipe2xi_publisher_channel = Pipe()
    pipe4xi_consumer_channel, pipe4xi_publisher_channel = Pipe()
    pipe8xi_consumer_channel, pipe8xi_publisher_channel = Pipe()
    # ---------------------------------------------------------------------------------------------
    # create three locks
    #   
    lock_xi2 = Lock()
    lock_xi4 = Lock()
    lock_xi8 = Lock()    
    # ---------------------------------------------------------------------------------------------
    # create three managers
    #       
    manager2xi = Manager()
    manager4xi = Manager()
    manager8xi = Manager()
    # ---------------------------------------------------------------------------------------------
    # create three dictionaries using each manager
    #      
    dict2xi = manager2xi.dict()
    dict4xi = manager4xi.dict()
    dict8xi = manager8xi.dict()
    
    # ---------------------------------------------------------------------------------------------
    # create three lists using each manager
    #      
    list2xi = manager2xi.list(range(15))
    list4xi = manager4xi.list(range(7))
    list8xi = manager8xi.list(range(5))
    
        
    # ---------------------------------------------------------------------------------------------
    # SENSE IT!
    #        create a process and use its services such the content it sends via its queue and pipe    
    #
    print("\n~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~2~")
    process2xi = Process(target=sense, args=('Energy ',queue2xi, pipe2xi_publisher_channel,lock_xi2,dict2xi,list2xi))
    process2xi.start()
    
    print ("process2xi alive status: ", process2xi.is_alive())       # check whether the process is alive
    print ("process2xi name: ", process2xi.name)                     # the name of the process
    print ("process2xi PID: ", process2xi.ident)                     # pid = ident of the process
    print ("process2xi authentication key: ", process2xi.authkey)    # authorization key given by the system
    
    print (queue2xi.get())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    
    print("\n Dictionary created by manager2xi and updated by process2xi:")
    print (dict2xi)
    
    print("\n List for manager2xi and process2xi:")
    print (list2xi)
    
    process2xi.join()
    print ("process2xi exit code after join : ", process2xi.exitcode) # see how the process did exit
    
    # ---------------------------------------------------------------------------------------------
    # REASON IT!
    #        create a process and use its services such the content it sends via its queue and pipe    
    # 
    print("\n~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~4~")
    process4xi = Process(target=reason, args=('Energy ',queue4xi, pipe4xi_publisher_channel,lock_xi4,dict4xi,list8xi))
    process4xi.start()
    
       
    print ("process4xi alive status: ", process4xi.is_alive())       # check whether the process is alive
    print ("process4xi name: ", process4xi.name)                     # the name of the process
    print ("process4xi PID: ", process4xi.ident)                     # pid = ident of the process
    print ("process4xi authentication key: ", process4xi.authkey)    # authorization key given by the system
    
    
    print (queue4xi.get())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    
    print("\n Dictionary created by manager4xi and updated by process4xi:")
    print (dict4xi)
    
    print("\n List for manager4xi and process4xi:")
    print (list4xi)
    
    process4xi.join()
    print ("process4xi exit code after join : ", process4xi.exitcode) # see how the process did exit
    
    # ---------------------------------------------------------------------------------------------
    # ACE IT!
    #        create a process and use its services such the content it sends via its queue and pipe    
    # 
    print("\n~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~8~")
    process8xi = Process(target=ace, args=('Energy ',queue8xi, pipe8xi_publisher_channel,lock_xi8,dict8xi,list8xi))
    process8xi.start()
    
           
    print ("process8xi alive status: ", process8xi.is_alive())       # check whether the process is alive
    print ("process8xi name: ", process8xi.name)                     # the name of the process
    print ("process8xi PID: ", process8xi.ident)                     # pid = ident of the process
    print ("process8xi authentication key: ", process8xi.authkey)    # authorization key given by the system
    
    
    print (queue8xi.get())
    print (pipe8xi_consumer_channel.recv())
    print (pipe8xi_consumer_channel.recv())
    print (pipe8xi_consumer_channel.recv())
    print (pipe8xi_consumer_channel.recv())
    print (pipe8xi_consumer_channel.recv())
    
    print("\n Dictionary created by manager8xi and updated by process8xi:")
    print (dict8xi)
    
    print("\n List for manager8xi and process8xi:")
    print (list8xi)
    
    process8xi.join()
    print ("process8xi exit code after join : ", process8xi.exitcode) # see how the process did exit
    
    # ---------------------------------------------------------------------------------------------
    # create three pools with worker processes 
    #        
    # pool2xi = Pool(processes=2)
    # pool4xi = Pool(processes=4)
    # pool8xi = Pool(processes=8)

    # result2xi = pool2xi.apply_async(expxi2, (2,))
    # print ("\nResult from pool2xi workers: ",result2xi.get(timeout=1))    
    
    # result4xi = pool4xi.apply_async(expxi4, (4,))
    # print ("\nResult from pool4xi workers: ",result4xi.get(timeout=1))
    
    # result8xi = pool8xi.apply_async(expxi8, (8,))
    # print ("\nResult from pool8xi workers: ",result8xi.get(timeout=1))
    
    # ---------------------------------------------------------------------------------------------
    # explore modifying ctypes objects in shared memory 
    #      
    lock_xi2 = Lock()

    n_xi2 = Value('i', 7)
    x_xi2 = Value(c_double, 1.0/3.0, lock=lock_xi2)
    s_xi2 = Array('c', 'yourDragonXi', lock=lock_xi2)
    A_xi2 = Array(Point2Xi, [(1.875,-6.25), (-5.75,2.0), (2.375,9.5)], lock=lock_xi2)

    # multiprocessing.Array(typecode_or_type, size_or_initializer, *, lock=True)
    #
    # Return a ctypes array allocated from shared memory. By default the return value is actually a synchronized wrapper for the array.
    # typecode_or_type
    #    determines the type of the elements of the returned array: it is either a ctypes type or a one character typecode of the kind used by the array module. 
    # size_or_initializer
    #    If size_or_initializer is an integer, then it determines the length of the array, and the array will be initially zeroed. 
    #    Otherwise, size_or_initializer is a sequence which is used to initialize the array and whose length determines the length of the array.
    # lock
    #    If lock is True (the default) then a new lock object is created to synchronize access to the value. 
    #    If lock is a Lock or RLock object then that will be used to synchronize access to the value. 
    #    If lock is False then access to the returned object will not be automatically protected by a lock, so it will not necessarily be process-safe
    #
    # Notes
    #    lock is a keyword only argument
    #    an array of ctypes.c_char has value and rawvalue attributes which allow one to use it to store and retrieve strings
    #

    process2ctypes = Process(target=modify2xi, args=(n_xi2, x_xi2, s_xi2, A_xi2))
    process2ctypes.start()
    process2ctypes.join()

    print ("n_xi2.value: ",n_xi2.value)
    print ("x_xi2.value: ",x_xi2.value)
    print ("s_xi2.value: ",s_xi2.value)
    print ([(a.x, a.y) for a in A_xi2])
    
    # ---------------------------------------------------------------------------------------------
    # explore with 3D coordinates
    
    lock_xi3 = Lock()
    n_xi3 = Value('i', 8)
    x_xi3 = Value(c_double, 1.0/3.0, lock=lock_xi3)
    s_xi3 = Array('c', 'DragonXi3D', lock=lock_xi3)
    A_xi3 = Array(ThreeDimension3Xi, [(1.875,-6.25,2.00), (-5.75,2.0,3.00), (2.375,9.5,4.0)], lock=lock_xi3)
    
    process3xyz = Process(target=modify3xi, args=(n_xi3, x_xi3, s_xi3, A_xi3))
    process3xyz.start()
    process3xyz.join()
    
    print ("n_xi3.value: ",n_xi3.value)
    print ("x_xi3.value: ",x_xi3.value)
    print ("s_xi3.value: ",s_xi3.value)
    print ([(a.x, a.y, a.z) for a in A_xi3])
    
       
    print("\n===============================================================")     

# -------------------------------------------------------------------------------------------------
# phase2: dragonized multiprocessing used                              
# -------------------------------------------------------------------------------------------------
def phase2():
    # ---------------------------------------------------------------------------------------------
    # create multiprocessing queues for sensing, reasoning and acing
    # 
    queue2xi = Queue2Xi()
    queue4xi = Queue2Xi()
    queue8xi = Queue2Xi()
    # ---------------------------------------------------------------------------------------------
    # create multiprocessing pipes with consumer and publisher ends
    #     
    pipe2xi_consumer_channel, pipe2xi_publisher_channel = Pipe2Xi()
    pipe4xi_consumer_channel, pipe4xi_publisher_channel = Pipe2Xi()
    pipe8xi_consumer_channel, pipe8xi_publisher_channel = Pipe2Xi()
    # ---------------------------------------------------------------------------------------------
    # create three locks
    #   
    lock_xi2 = Lock2Xi()
    lock_xi4 = Lock2Xi()
    lock_xi8 = Lock2Xi()    
    # ---------------------------------------------------------------------------------------------
    # create three managers
    #       
    manager2xi = Manager2Xi()
    manager4xi = Manager2Xi()
    manager8xi = Manager2Xi()
    # ---------------------------------------------------------------------------------------------
    # create three dictionaries using each manager
    #      
    dict2xi = manager2xi.dict()
    dict4xi = manager4xi.dict()
    dict8xi = manager8xi.dict()
    
    # ---------------------------------------------------------------------------------------------
    # create three lists using each manager
    #      
    list2xi = manager2xi.list(range(15))
    list4xi = manager4xi.list(range(7))
    list8xi = manager8xi.list(range(5))
    
        
    # ---------------------------------------------------------------------------------------------
    # SENSE IT!
    #        create a process and use its services such the content it sends via its queue and pipe    
    #
    print("\n SENSE with yourDragonXi!")
    process2xi = Process2Xi(target=sense, args=('Energy ',queue2xi, pipe2xi_publisher_channel,lock_xi2,dict2xi,list2xi))
    process2xi.start()
    
    print ("process2xi alive status: ", process2xi.is_alive())       # check whether the process is alive
    print ("process2xi name: ", process2xi.name)                     # the name of the process
    print ("process2xi PID: ", process2xi.ident)                     # pid = ident of the process
    print ("process2xi authentication key: ", process2xi.authkey)    # authorization key given by the system
    
    print (queue2xi.get())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    print (pipe2xi_consumer_channel.recv())
    
    print("\n Dictionary created by manager2xi and updated by process2xi:")
    print (dict2xi)
    
    print("\n List for manager2xi and process2xi:")
    print (list2xi)
    
    process2xi.join()
    print ("process2xi exit code after join : ", process2xi.exitcode) # see how the process did exit
    
    # ---------------------------------------------------------------------------------------------
    # REASON IT!
    #        create a process and use its services such the content it sends via its queue and pipe    
    # 
    print("\n REASON with yourDragonXi")
    process4xi = Process2Xi(target=reason, args=('Energy ',queue4xi, pipe4xi_publisher_channel,lock_xi4,dict4xi,list8xi))
    process4xi.start()
    
       
    print ("process4xi alive status: ", process4xi.is_alive())       # check whether the process is alive
    print ("process4xi name: ", process4xi.name)                     # the name of the process
    print ("process4xi PID: ", process4xi.ident)                     # pid = ident of the process
    print ("process4xi authentication key: ", process4xi.authkey)    # authorization key given by the system
    
    
    print (queue4xi.get())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    print (pipe4xi_consumer_channel.recv())
    
    print("\n Dictionary created by manager4xi and updated by process4xi:")
    print (dict4xi)
    
    print("\n List for manager4xi and process4xi:")
    print (list4xi)
    
    process4xi.join()
    print ("process4xi exit code after join : ", process4xi.exitcode) # see how the process did exit
    
    # ---------------------------------------------------------------------------------------------
    # ACE IT!
    #        create a process and use its services such the content it sends via its queue and pipe    
    # 
    print("\n ACE with yourDragonXi!")
    process8xi = Process2Xi(target=ace, args=('Energy ',queue8xi, pipe8xi_publisher_channel,lock_xi8,dict8xi,list8xi))
    process8xi.start()
    
           
    print ("process8xi alive status: ", process8xi.is_alive())       # check whether the process is alive
    print ("process8xi name: ", process8xi.name)                     # the name of the process
    print ("process8xi PID: ", process8xi.ident)                     # pid = ident of the process
    print ("process8xi authentication key: ", process8xi.authkey)    # authorization key given by the system
    
    
    print (queue8xi.get())
    print (pipe8xi_consumer_channel.recv())
    print (pipe8xi_consumer_channel.recv())
    print (pipe8xi_consumer_channel.recv())
    print (pipe8xi_consumer_channel.recv())
    print (pipe8xi_consumer_channel.recv())
    
    print("\n Dictionary created by manager8xi and updated by process8xi:")
    print (dict8xi)
    
    print("\n List for manager8xi and process8xi:")
    print (list8xi)
    
    process8xi.join()
    print ("process8xi exit code after join : ", process8xi.exitcode) # see how the process did exit
    
    # ---------------------------------------------------------------------------------------------
    # create three pools with worker processes 
    #        
    # pool2xi = Pool(processes=2)
    # pool4xi = Pool(processes=4)
    # pool8xi = Pool(processes=8)

    # result2xi = pool2xi.apply_async(expxi2, (2,))
    # print ("\nResult from pool2xi workers: ",result2xi.get(timeout=1))    
    
    # result4xi = pool4xi.apply_async(expxi4, (4,))
    # print ("\nResult from pool4xi workers: ",result4xi.get(timeout=1))
    
    # result8xi = pool8xi.apply_async(expxi8, (8,))
    # print ("\nResult from pool8xi workers: ",result8xi.get(timeout=1))
    
    # ---------------------------------------------------------------------------------------------
    # explore modifying ctypes objects in shared memory 
    #      
    lock_xi2 = Lock2Xi()

    n_xi2 = Value2Xi('i', 7)
    x_xi2 = Value2Xi(c_double, 1.0/3.0, lock=lock_xi2)
    s_xi2 = ArrayFunction2Xi('c', 'yourDragonXi', lock=lock_xi2)
    A_xi2 = ArrayFunction2Xi(Point2Xi, [(1.875,-6.25), (-5.75,2.0), (2.375,9.5)], lock=lock_xi2)

    # multiprocessing.Array(typecode_or_type, size_or_initializer, *, lock=True)
    #
    # Return a ctypes array allocated from shared memory. By default the return value is actually a synchronized wrapper for the array.
    # typecode_or_type
    #    determines the type of the elements of the returned array: it is either a ctypes type or a one character typecode of the kind used by the array module. 
    # size_or_initializer
    #    If size_or_initializer is an integer, then it determines the length of the array, and the array will be initially zeroed. 
    #    Otherwise, size_or_initializer is a sequence which is used to initialize the array and whose length determines the length of the array.
    # lock
    #    If lock is True (the default) then a new lock object is created to synchronize access to the value. 
    #    If lock is a Lock or RLock object then that will be used to synchronize access to the value. 
    #    If lock is False then access to the returned object will not be automatically protected by a lock, so it will not necessarily be process-safe
    #
    # Notes
    #    lock is a keyword only argument
    #    an array of ctypes.c_char has value and rawvalue attributes which allow one to use it to store and retrieve strings
    #

    process2ctypes = Process2Xi(target=modify2xi, args=(n_xi2, x_xi2, s_xi2, A_xi2))
    process2ctypes.start()
    process2ctypes.join()

    print ("n_xi2.value: ",n_xi2.value)
    print ("x_xi2.value: ",x_xi2.value)
    print ("s_xi2.value: ",s_xi2.value)
    print ([(a.x, a.y) for a in A_xi2])
    
    # ---------------------------------------------------------------------------------------------
    # explore with 3D coordinates
    
    lock_xi3 = Lock2Xi()
    n_xi3 = Value2Xi('i', 8)
    x_xi3 = Value2Xi(c_double, 1.0/3.0, lock=lock_xi3)
    
    # ---------------------------------------------------------------------------------------------
    # note: in this case we have to call ArrayFunction2Xi defined in arrayfunction2xi.py
    #       i.e. NOT Array2Xi, whic is a class defined by array2xi.py
        
    s_xi3 = ArrayFunction2Xi('c', 'DragonXi3D', lock=lock_xi3)
    A_xi3 = ArrayFunction2Xi(ThreeDimension3Xi, [(1.875,-6.25,2.00), (-5.75,2.0,3.00), (2.375,9.5,4.0)], lock=lock_xi3)
    
    process3xyz = Process2Xi(target=modify3xi, args=(n_xi3, x_xi3, s_xi3, A_xi3))
    process3xyz.start()
    process3xyz.join()
    
    print ("n_xi3.value: ",n_xi3.value)
    print ("x_xi3.value: ",x_xi3.value)
    print ("s_xi3.value: ",s_xi3.value)
    print ([(a.x, a.y, a.z) for a in A_xi3])
    
       
    print("\n===============================================================") 
    

# -------------------------------------------------------------------------------------------------
# main function                              
# -------------------------------------------------------------------------------------------------
def main():

    phase1()
    
    phase2()
    
# -------------------------------------------------------------------------------------------------
#                               
# -------------------------------------------------------------------------------------------------
if __name__ == '__main__':
    main()


# -------------------------------------------------------------------------------------------------
# phase2 output                                
# -------------------------------------------------------------------------------------------------
#  SENSE with yourDragonXi!
# process2xi alive status:  True
# process2xi name:  Process2Xi-4
# process2xi PID:  3936
# process2xi authentication key:  b'BY\xceM\x1a\x8eu\xa6\xa0\xfa\xa7X~\x98\x8d\x04\xb40\x9a\xfda\\8@\x1d\xe4\x9byf`\\\xbb'
# ['DragonXi2_00', 'Energy ', 'sensing!']
# ['DragonXi2_Sensing_01', 'BIO ', 'Energy ']
# ['DragonXi2_Sensing_02', 'CO2 ', 'Energy ']
# ['DragonXi2_Sensing_03', 'Coal ', 'Energy ']
# ['DragonXi2_Sensing_04', 'Diesel ', 'Energy ']
# ['DragonXi2_Sensing_05', 'Electricity ', 'Energy ']
# ['DragonXi2_Sensing_06', 'Fusion ', 'Energy ']
# ['DragonXi2_Sensing_07', 'Gas ', 'Energy ']
# ['DragonXi2_Sensing_08', 'Geo ', 'Energy ']
# ['DragonXi2_Sensing_09', 'Hydrogen ', 'Energy ']
# ['DragonXi2_Sensing_10', 'Lithium ', 'Energy ']
# ['DragonXi2_Sensing_11', 'Nuclear Power ', 'Energy ']
# ['DragonXi2_Sensing_12', 'Ocean ', 'Energy ']
# ['DragonXi2_Sensing_13', 'Oil ', 'Energy ']
# ['DragonXi2_Sensing_14', 'Solar ', 'Energy ']
# ['DragonXi2_Sensing_15', 'Wind ', 'Energy ']
# 
#  Dictionary created by manager2xi and updated by process2xi:
# {1: 'BIO', 2: 'CO2', 3: 'Coal', 4: 'Diesel', 5: 'Electricity'}
# 
#  List for manager2xi and process2xi:
# [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
# process2xi exit code after join :  0
# 
#  REASON with yourDragonXi
# process4xi alive status:  True
# process4xi name:  Process2Xi-5
# process4xi PID:  3016
# process4xi authentication key:  b'BY\xceM\x1a\x8eu\xa6\xa0\xfa\xa7X~\x98\x8d\x04\xb40\x9a\xfda\\8@\x1d\xe4\x9byf`\\\xbb'
# ['DragonXi4_00', 'Energy ', 'portfolio must be sustainable!']
# ['DragonXi4_Reasoning_01', 'BIO ', 'Energy ']
# ['DragonXi4_Reasoning_02', 'Geo ', 'Energy ']
# ['DragonXi4_Reasoning_03', 'Hydrogen ', 'Energy ']
# ['DragonXi4_Reasoning_04', 'Nuclear Power ', 'Energy ']
# ['DragonXi4_Reasoning_05', 'Ocean ', 'Energy ']
# ['DragonXi4_Reasoning_06', 'Solar ', 'Energy ']
# ['DragonXi4_Reasoning_07', 'Wind ', 'Energy ']
# 
#  Dictionary created by manager4xi and updated by process4xi:
# {1: 'BIO', 2: 'Geo', 3: 'Hydrogen', 4: 'Nuclear Power', 5: 'Ocean', 6: 'Solar'}
# 
#  List for manager4xi and process4xi:
# [0, 1, 2, 3, 4, 5, 6]
# process4xi exit code after join :  0
# 
#  ACE with yourDragonXi!
# process8xi alive status:  True
# process8xi name:  Process2Xi-6
# process8xi PID:  3456
# process8xi authentication key:  b'BY\xceM\x1a\x8eu\xa6\xa0\xfa\xa7X~\x98\x8d\x04\xb40\x9a\xfda\\8@\x1d\xe4\x9byf`\\\xbb'
# ['DragonXi8_00', 'Energy ', 'to be aced!']
# ['DragonXi8_Ace_01', 'BIO ', 'Energy ']
# ['DragonXi8_Ace_02', 'Hydrogen ', 'Energy ']
# ['DragonXi8_Ace_03', 'Nuclear Power ', 'Energy ']
# ['DragonXi8_Ace_04', 'Solar ', 'Energy ']
# ['DragonXi8_Ace_05', 'Wind ', 'Energy ']
# 
#  Dictionary created by manager8xi and updated by process8xi:
# {1: 'BIO', 2: 'Hydrogen'}
# 
#  List for manager8xi and process8xi:
# [0, 1, 2, 3, 4]
# process8xi exit code after join :  0
# n_xi2.value:  7
# x_xi2.value:  0.333333333333
# s_xi2.value:  b'YOURDRAGONXI'
# [(3.515625, 39.0625), (33.0625, 4.0), (5.640625, 90.25)]
# n_xi3.value:  8
# x_xi3.value:  0.333333333333
# s_xi3.value:  b'DRAGONXI3D'
# [(6.591796875, -244.140625, 8.0), (-190.109375, 8.0, 27.0), (13.396484375, 857.375, 64.0)]
# 
# -------------------------------------------------------------------------------------------------
# end of test_ctypes_mp_xi2.py                             
# -------------------------------------------------------------------------------------------------